@@ -11,7 +11,7 @@
 // This module provides a lot of the trivial WavPack API functions and several
 // functions that are common to both reading and writing WavPack files (like
 // WavpackCloseFile()). Functions here are restricted to those that have few
-// external dependancies and this is done so that applications that statically
+// external dependencies and this is done so that applications that statically
 // link to the WavPack library (like the command-line utilities on Windows)
 // do not need to include the entire library image if they only use a subset
 // of it. This module will be loaded for ANY WavPack application.
@@ -40,7 +40,7 @@ const uint32_t sample_rates [] = { 6000, 8000, 9600, 11025, 12000, 16000, 22050,
 // MODE_LOSSLESS:  file is lossless (either pure or hybrid)
 // MODE_HYBRID:  file is hybrid mode (either lossy or lossless)
 // MODE_FLOAT:  audio data is 32-bit ieee floating point
-// MODE_VALID_TAG:  file conatins a valid ID3v1 or APEv2 tag
+// MODE_VALID_TAG:  file contains a valid ID3v1 or APEv2 tag
 // MODE_HIGH:  file was created in "high" mode (information only)
 // MODE_FAST:  file was created in "fast" mode (information only)
 // MODE_EXTRA:  file was created using "extra" mode (information only)
@@ -103,11 +103,37 @@ int WavpackGetMode (WavpackContext *wpc)
                 mode |= MODE_APETAG;
         }
 #endif
+
+        mode |= (wpc->config.qmode << 16) & 0xFF0000;
     }
 
     return mode;
 }
 
+// This function obtains information about specific file features that were
+// added for version 5.0, specifically qualifications added to support CAF
+// and DSD files. Except for indicating the presence of DSD data, these
+// bits are meant to simply indicate the format of the data in the original
+// source file and do NOT indicate how the library will return the data to
+// the appication (which is always the same). This means that in general an
+// application that simply wants to play or process the audio data need not
+// be concerned about these. If the file is DSD audio, then either of the
+// QMDOE_DSD_LSB_FIRST or QMODE_DSD_MSB_FIRST bits will be set (but the
+// DSD audio is always returned to the caller MSB first).
+
+// QMODE_BIG_ENDIAN        0x1     // big-endian data format (opposite of WAV format)
+// QMODE_SIGNED_BYTES      0x2     // 8-bit audio data is signed (opposite of WAV format)
+// QMODE_UNSIGNED_WORDS    0x4     // audio data (other than 8-bit) is unsigned (opposite of WAV format)
+// QMODE_REORDERED_CHANS   0x8     // source channels were not Microsoft order, so they were reordered
+// QMODE_DSD_LSB_FIRST     0x10    // DSD bytes, LSB first (most Sony .dsf files)
+// QMODE_DSD_MSB_FIRST     0x20    // DSD bytes, MSB first (Philips .dff files)
+// QMODE_DSD_IN_BLOCKS     0x40    // DSD data is blocked by channels (Sony .dsf only)
+
+int WavpackGetQualifyMode (WavpackContext *wpc)
+{
+    return wpc->config.qmode & 0xFF;
+}
+
 // This function returns a pointer to a string describing the last error
 // generated by WavPack.
 
@@ -120,15 +146,25 @@ char *WavpackGetErrorMessage (WavpackContext *wpc)
 
 uint32_t WavpackGetNumSamples (WavpackContext *wpc)
 {
-    return wpc ? wpc->total_samples : (uint32_t) -1;
+    return (uint32_t) WavpackGetNumSamples64 (wpc);
+}
+
+int64_t WavpackGetNumSamples64 (WavpackContext *wpc)
+{
+    return wpc ? wpc->total_samples : -1;
 }
 
 // Get the current sample index position, or -1 if unknown
 
 uint32_t WavpackGetSampleIndex (WavpackContext *wpc)
+{
+    return (uint32_t) WavpackGetSampleIndex64 (wpc);
+}
+
+int64_t WavpackGetSampleIndex64 (WavpackContext *wpc)
 {
     if (wpc) {
-#ifndef VER4_ONLY
+#ifdef ENABLE_LEGACY
         if (wpc->stream3)
             return get_sample_index3 (wpc);
         else if (wpc->streams && wpc->streams [0])
@@ -139,7 +175,7 @@ uint32_t WavpackGetSampleIndex (WavpackContext *wpc)
 #endif
     }
 
-    return (uint32_t) -1;
+    return -1;
 }
 
 // Get the number of errors encountered so far
@@ -162,8 +198,8 @@ int WavpackLossyBlocks (WavpackContext *wpc)
 
 double WavpackGetProgress (WavpackContext *wpc)
 {
-    if (wpc && wpc->total_samples != (uint32_t) -1 && wpc->total_samples != 0)
-        return (double) WavpackGetSampleIndex (wpc) / wpc->total_samples;
+    if (wpc && wpc->total_samples != -1 && wpc->total_samples != 0)
+        return (double) WavpackGetSampleIndex64 (wpc) / wpc->total_samples;
     else
         return -1.0;
 }
@@ -171,6 +207,11 @@ double WavpackGetProgress (WavpackContext *wpc)
 // Return the total size of the WavPack file(s) in bytes.
 
 uint32_t WavpackGetFileSize (WavpackContext *wpc)
+{
+    return (uint32_t) (wpc ? wpc->filelen + wpc->file2len : 0);
+}
+
+int64_t WavpackGetFileSize64 (WavpackContext *wpc)
 {
     return wpc ? wpc->filelen + wpc->file2len : 0;
 }
@@ -182,7 +223,7 @@ uint32_t WavpackGetFileSize (WavpackContext *wpc)
 
 double WavpackGetRatio (WavpackContext *wpc)
 {
-    if (wpc && wpc->total_samples != (uint32_t) -1 && wpc->filelen) {
+    if (wpc && wpc->total_samples != -1 && wpc->filelen) {
         double output_size = (double) wpc->total_samples * wpc->config.num_channels *
             wpc->config.bytes_per_sample;
         double input_size = (double) wpc->filelen + wpc->file2len;
@@ -200,8 +241,8 @@ double WavpackGetRatio (WavpackContext *wpc)
 
 double WavpackGetAverageBitrate (WavpackContext *wpc, int count_wvc)
 {
-    if (wpc && wpc->total_samples != (uint32_t) -1 && wpc->filelen) {
-        double output_time = (double) wpc->total_samples / wpc->config.sample_rate;
+    if (wpc && wpc->total_samples != -1 && wpc->filelen && WavpackGetSampleRate (wpc)) {
+        double output_time = (double) wpc->total_samples / WavpackGetSampleRate (wpc);
         double input_size = (double) wpc->filelen + (count_wvc ? wpc->file2len : 0);
 
         if (output_time >= 0.1 && input_size >= 1.0)
@@ -221,8 +262,8 @@ double WavpackGetInstantBitrate (WavpackContext *wpc)
     if (wpc && wpc->stream3)
         return WavpackGetAverageBitrate (wpc, TRUE);
 
-    if (wpc && wpc->streams && wpc->streams [0] && wpc->streams [0]->wphdr.block_samples) {
-        double output_time = (double) wpc->streams [0]->wphdr.block_samples / wpc->config.sample_rate;
+    if (wpc && wpc->streams && wpc->streams [0] && wpc->streams [0]->wphdr.block_samples && WavpackGetSampleRate (wpc)) {
+        double output_time = (double) wpc->streams [0]->wphdr.block_samples / WavpackGetSampleRate (wpc);
         double input_size = 0;
         int si;
 
@@ -241,11 +282,86 @@ double WavpackGetInstantBitrate (WavpackContext *wpc)
     return 0.0;
 }
 
+// This function allows retrieving the Core Audio File channel layout, many of which do not
+// conform to the Microsoft ordering standard that WavPack requires internally (at least for
+// those channels present in the "channel mask"). In addition to the layout tag, this function
+// returns the reordering string (if stored in the file) to allow the unpacker to reorder the
+// channels back to the specified layout (if it wants to restore the CAF order). The number of
+// channels in the layout is determined from the lower nybble of the layout word (and should
+// probably match the number of channels in the file), and if a reorder string is requested
+// then that much space must be allocated. Note that all the reordering is actually done
+// outside of this library, and that if reordering is done then the appropriate qmode bit
+// will be set.
+//
+// Note: Normally this function would not be used by an application unless it specifically
+// wanted to restore a non-standard channel order (to check an MD5, for example) or obtain
+// the Core Audio channel layout ID. For simple file decoding for playback, the channel_mask
+// should provide all the information required unless there are non-Microsoft channels
+// involved, in which case WavpackGetChannelIdentities() will provide the identities of
+// the other channels (if they are known).
+
+uint32_t WavpackGetChannelLayout (WavpackContext *wpc, unsigned char *reorder)
+{
+    if ((wpc->channel_layout & 0xff) && wpc->channel_reordering && reorder)
+        memcpy (reorder, wpc->channel_reordering, wpc->channel_layout & 0xff);
+
+    return wpc->channel_layout;
+}
+
+// This function provides the identities of ALL the channels in the file, including the
+// standard Microsoft channels (which come first, in order, and are numbered 1-18) and also
+// any non-Microsoft channels (which can be in any order and have values from 33-254). The
+// value 0x00 is invalid and 0xFF indicates an "unknown" or "unnassigned" channel. The
+// string is NULL terminated so the caller must supply enough space for the number
+// of channels indicated by WavpackGetNumChannels(), plus one.
+//
+// Note that this function returns the actual order of the channels in the Wavpack file
+// (i.e., the order returned by WavpackUnpackSamples()). If the file includes a "reordering"
+// string because the source file was not in Microsoft order that is NOT taken into account
+// here and really only needs to be considered if doing an MD5 verification or if it's
+// required to restore the original order/file (like wvunpack does).
+
+void WavpackGetChannelIdentities (WavpackContext *wpc, unsigned char *identities)
+{
+    int num_channels = wpc->config.num_channels, index = 1;
+    uint32_t channel_mask = wpc->config.channel_mask;
+    unsigned char *src = wpc->channel_identities;
+
+    while (num_channels--) {
+        if (channel_mask) {
+            while (!(channel_mask & 1)) {
+                channel_mask >>= 1;
+                index++;
+            }
+
+            *identities++ = index++;
+            channel_mask >>= 1;
+        }
+        else if (src && *src)
+            *identities++ = *src++;
+        else
+            *identities++ = 0xff;
+    }
+
+    *identities = 0;
+}
+
+// For local use only. Install a callback to be executed when WavpackCloseFile() is called,
+// usually used to dump some statistics accumulated during encode or decode.
+
+void install_close_callback (WavpackContext *wpc, void cb_func (void *wpc))
+{
+    wpc->close_callback = cb_func;
+}
+
 // Close the specified WavPack file and release all resources used by it.
 // Returns NULL.
 
 WavpackContext *WavpackCloseFile (WavpackContext *wpc)
 {
+    if (wpc->close_callback)
+        wpc->close_callback (wpc);
+
     if (wpc->streams) {
         free_streams (wpc);
 
@@ -255,27 +371,44 @@ WavpackContext *WavpackCloseFile (WavpackContext *wpc)
         free (wpc->streams);
     }
 
-#ifndef VER4_ONLY
+#ifdef ENABLE_LEGACY
     if (wpc->stream3)
         free_stream3 (wpc);
 #endif
 
-    if (wpc->close_files) {
-#ifndef NO_USE_FSTREAMS
-        if (wpc->wv_in != NULL)
-            fclose (wpc->wv_in);
+    if (wpc->reader && wpc->reader->close && wpc->wv_in)
+        wpc->reader->close (wpc->wv_in);
 
-        if (wpc->wvc_in != NULL)
-            fclose (wpc->wvc_in);
-#endif
-    }
+    if (wpc->reader && wpc->reader->close && wpc->wvc_in)
+        wpc->reader->close (wpc->wvc_in);
 
     WavpackFreeWrapper (wpc);
 
+    if (wpc->metadata) {
+        int i;
+
+        for (i = 0; i < wpc->metacount; ++i)
+            if (wpc->metadata [i].data)
+                free (wpc->metadata [i].data);
+
+        free (wpc->metadata);
+    }
+
+    if (wpc->channel_identities)
+        free (wpc->channel_identities);
+
+    if (wpc->channel_reordering)
+        free (wpc->channel_reordering);
+
 #ifndef NO_TAGS
     free_tag (&wpc->m_tag);
 #endif
 
+#ifdef ENABLE_DSD
+    if (wpc->decimation_context)
+        decimate_dsd_destroy (wpc->decimation_context);
+#endif
+
     free (wpc);
 
     return NULL;
@@ -309,7 +442,16 @@ void WavpackFreeWrapper (WavpackContext *wpc)
 
 uint32_t WavpackGetSampleRate (WavpackContext *wpc)
 {
-    return wpc ? wpc->config.sample_rate : 44100;
+    return wpc ? (wpc->dsd_multiplier ? wpc->config.sample_rate * wpc->dsd_multiplier : wpc->config.sample_rate) : 44100;
+}
+
+// Returns the native sample rate of the specified WavPack file
+// (provides the native rate for DSD files rather than the "byte" rate that's used for
+//   seeking, duration, etc. and would generally be used just for user facing reports)
+
+uint32_t WavpackGetNativeSampleRate (WavpackContext *wpc)
+{
+    return wpc ? (wpc->dsd_multiplier ? wpc->config.sample_rate * wpc->dsd_multiplier * 8 : wpc->config.sample_rate) : 44100;
 }
 
 // Returns the number of channels of the specified WavPack file. Note that
@@ -378,7 +520,7 @@ int WavpackGetReducedChannels (WavpackContext *wpc)
 }
 
 // Free all memory allocated for raw WavPack blocks (for all allocated streams)
-// and free all additonal streams. This does not free the default stream ([0])
+// and free all additional streams. This does not free the default stream ([0])
 // which is always kept around.
 
 void free_streams (WavpackContext *wpc)
@@ -406,6 +548,10 @@ void free_streams (WavpackContext *wpc)
             wpc->streams [si]->dc.shaping_data = NULL;
         }
 
+#ifdef ENABLE_DSD
+        free_dsd_tables (wpc->streams [si]);
+#endif
+
         if (si) {
             wpc->num_streams--;
             free (wpc->streams [si]);
@@ -416,6 +562,34 @@ void free_streams (WavpackContext *wpc)
     wpc->current_stream = 0;
 }
 
+void free_dsd_tables (WavpackStream *wps)
+{
+    if (wps->dsd.probabilities) {
+        free (wps->dsd.probabilities);
+        wps->dsd.probabilities = NULL;
+    }
+
+    if (wps->dsd.summed_probabilities) {
+        free (wps->dsd.summed_probabilities);
+        wps->dsd.summed_probabilities = NULL;
+    }
+
+    if (wps->dsd.lookup_buffer) {
+        free (wps->dsd.lookup_buffer);
+        wps->dsd.lookup_buffer = NULL;
+    }
+
+    if (wps->dsd.value_lookup) {
+        free (wps->dsd.value_lookup);
+        wps->dsd.value_lookup = NULL;
+    }
+
+    if (wps->dsd.ptable) {
+        free (wps->dsd.ptable);
+        wps->dsd.ptable = NULL;
+    }
+}
+
 void WavpackFloatNormalize (int32_t *values, int32_t num_values, int delta_exp)
 {
     f32 *fvalues = (f32 *) values;
@@ -441,19 +615,26 @@ void WavpackFloatNormalize (int32_t *values, int32_t num_values, int delta_exp)
 void WavpackLittleEndianToNative (void *data, char *format)
 {
     unsigned char *cp = (unsigned char *) data;
-    int32_t temp;
+    int64_t temp;
 
     while (*format) {
         switch (*format) {
+            case 'D':
+                temp = cp [0] + ((int64_t) cp [1] << 8) + ((int64_t) cp [2] << 16) + ((int64_t) cp [3] << 24) +
+                    ((int64_t) cp [4] << 32) + ((int64_t) cp [5] << 40) + ((int64_t) cp [6] << 48) + ((uint64_t) cp [7] << 56);
+                * (int64_t *) cp = temp;
+                cp += 8;
+                break;
+
             case 'L':
-                temp = cp [0] + ((int32_t) cp [1] << 8) + ((int32_t) cp [2] << 16) + ((int32_t) cp [3] << 24);
-                * (int32_t *) cp = temp;
+                temp = cp [0] + ((int32_t) cp [1] << 8) + ((int32_t) cp [2] << 16) + ((int64_t) cp [3] << 24);
+                * (int32_t *) cp = (int32_t) temp;
                 cp += 4;
                 break;
 
             case 'S':
                 temp = cp [0] + (cp [1] << 8);
-                * (short *) cp = (short) temp;
+                * (int16_t *) cp = (int16_t) temp;
                 cp += 2;
                 break;
 
@@ -471,10 +652,22 @@ void WavpackLittleEndianToNative (void *data, char *format)
 void WavpackNativeToLittleEndian (void *data, char *format)
 {
     unsigned char *cp = (unsigned char *) data;
-    int32_t temp;
+    int64_t temp;
 
     while (*format) {
         switch (*format) {
+            case 'D':
+                temp = * (int64_t *) cp;
+                *cp++ = (unsigned char) temp;
+                *cp++ = (unsigned char) (temp >> 8);
+                *cp++ = (unsigned char) (temp >> 16);
+                *cp++ = (unsigned char) (temp >> 24);
+                *cp++ = (unsigned char) (temp >> 32);
+                *cp++ = (unsigned char) (temp >> 40);
+                *cp++ = (unsigned char) (temp >> 48);
+                *cp++ = (unsigned char) (temp >> 56);
+                break;
+
             case 'L':
                 temp = * (int32_t *) cp;
                 *cp++ = (unsigned char) temp;
@@ -484,7 +677,7 @@ void WavpackNativeToLittleEndian (void *data, char *format)
                 break;
 
             case 'S':
-                temp = * (short *) cp;
+                temp = * (int16_t *) cp;
                 *cp++ = (unsigned char) temp;
                 *cp++ = (unsigned char) (temp >> 8);
                 break;
@@ -500,6 +693,87 @@ void WavpackNativeToLittleEndian (void *data, char *format)
     }
 }
 
+void WavpackBigEndianToNative (void *data, char *format)
+{
+    unsigned char *cp = (unsigned char *) data;
+    int64_t temp;
+
+    while (*format) {
+        switch (*format) {
+            case 'D':
+                temp = cp [7] + ((int64_t) cp [6] << 8) + ((int64_t) cp [5] << 16) + ((int64_t) cp [4] << 24) +
+                    ((int64_t) cp [3] << 32) + ((int64_t) cp [2] << 40) + ((int64_t) cp [1] << 48) + ((uint64_t) cp [0] << 56);
+                * (int64_t *) cp = temp;
+                cp += 8;
+                break;
+
+            case 'L':
+                temp = cp [3] + ((int32_t) cp [2] << 8) + ((int32_t) cp [1] << 16) + ((int64_t) cp [0] << 24);
+                * (int32_t *) cp = (int32_t) temp;
+                cp += 4;
+                break;
+
+            case 'S':
+                temp = cp [1] + (cp [0] << 8);
+                * (int16_t *) cp = (int16_t) temp;
+                cp += 2;
+                break;
+
+            default:
+                if (isdigit (*format))
+                    cp += *format - '0';
+
+                break;
+        }
+
+        format++;
+    }
+}
+
+void WavpackNativeToBigEndian (void *data, char *format)
+{
+    unsigned char *cp = (unsigned char *) data;
+    int64_t temp;
+
+    while (*format) {
+        switch (*format) {
+            case 'D':
+                temp = * (int64_t *) cp;
+                *cp++ = (unsigned char) (temp >> 56);
+                *cp++ = (unsigned char) (temp >> 48);
+                *cp++ = (unsigned char) (temp >> 40);
+                *cp++ = (unsigned char) (temp >> 32);
+                *cp++ = (unsigned char) (temp >> 24);
+                *cp++ = (unsigned char) (temp >> 16);
+                *cp++ = (unsigned char) (temp >> 8);
+                *cp++ = (unsigned char) temp;
+                break;
+
+            case 'L':
+                temp = * (int32_t *) cp;
+                *cp++ = (unsigned char) (temp >> 24);
+                *cp++ = (unsigned char) (temp >> 16);
+                *cp++ = (unsigned char) (temp >> 8);
+                *cp++ = (unsigned char) temp;
+                break;
+
+            case 'S':
+                temp = * (int16_t *) cp;
+                *cp++ = (unsigned char) (temp >> 8);
+                *cp++ = (unsigned char) temp;
+                break;
+
+            default:
+                if (isdigit (*format))
+                    cp += *format - '0';
+
+                break;
+        }
+
+        format++;
+    }
+}
+
 uint32_t WavpackGetLibraryVersion (void)
 {
     return (LIBWAVPACK_MAJOR<<16)
